#include <linux/init.h>
#include <linux/kernel.h>
#include <asm/bootinfo.h>
#include <linux/interrupt.h>
#include <asm/time.h>
#include <asm/traps.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>

#define GPIO_SPI_CLK   40
#define GPIO_SPI_MISO  63
#define GPIO_SPI_MOSI  49
#define IRQ_GPIO       6
#define CHIPSEL_GPIO   71

struct spi_gpio_platform_data {
	unsigned        sck;
	unsigned        mosi;
	unsigned        miso;
	u16             num_chipselect;
};

struct mcp251x_platform_data {
	unsigned long oscillator_frequency;
	int model;
#define CAN_MCP251X_MCP2510 0
#define CAN_MCP251X_MCP2515 1
	int (*board_specific_setup)(struct spi_device *spi);
	int (*transceiver_enable)(int enable);
	int (*power_enable) (int enable);
};

struct spi_gpio_platform_data spi_gpio_platform_data = {
    .sck = GPIO_SPI_CLK,
    .mosi = GPIO_SPI_MOSI,
    .miso = GPIO_SPI_MISO,
    .num_chipselect = 1,
};

static struct platform_device spi_gpio_device = {
    .name = "spi_gpio",
    .id = 2,
    .dev = {
        .platform_data = &spi_gpio_platform_data,
    }
};

static struct platform_device lmlsfpga_spi_device1 = {
    .name = "lmlsfpga_spi",
    .id = 3,
};

static struct platform_device lmlsfpga_spi_device2 = {
    .name = "lmlsfpga_spi",
    .id = 4,
};

static struct platform_device lmlsfpga_spi_device3 = {
    .name = "lmlsfpga_spi",
    .id = 5,
};

static struct platform_device lmlsfpga_spi_device4 = {
    .name = "lmlsfpga_spi",
    .id = 6,
};

static struct mcp251x_platform_data mcp251x_info = {
         .oscillator_frequency = 20000000,
         .board_specific_setup = NULL,
         .model = CAN_MCP251X_MCP2515,
         .power_enable = NULL, 
         .transceiver_enable = NULL,
};

static struct spi_board_info spi_gpio_mcp2515_info[] = {
    {
        .modalias = "mcp251x",
        .platform_data = &mcp251x_info,
        .irq = IRQ_GPIO, //GPIO_50 on Pin 6 for IRQ
        .max_speed_hz = 1*1000*1000,
        .chip_select = 0,
        .bus_num = 3,
        .mode = SPI_MODE_0,
    },
    {
        .modalias = "mcp251x",
        .platform_data = &mcp251x_info,
        .bus_num = 4,
        .mode = SPI_MODE_0,
    },
    {
        .modalias = "mcp251x",
        .platform_data = &mcp251x_info,
        .bus_num = 5,
        .mode = SPI_MODE_0,
    },
    {
        .modalias = "mcp251x",
        .platform_data = &mcp251x_info,
        .bus_num = 6,
        .mode = SPI_MODE_0,
    },
};

extern void mips_reboot_setup(void);
extern void mips_timer_setup(struct irqaction *irq);
static void vsc452_be_init(void);

#ifdef CONFIG_KGDB
extern void kgdb_config(void);
#endif

const char *get_system_type(void)
{
	return "VSC452";
}

void __init vsc452TimeInit(void)
{
	// clocked @ 300MHz, counter counts every other clock
	mips_hpt_frequency = 300e6/2;
}
static void vsc452_be_init(void)
{
}

int vsc452_be_handler (struct pt_regs *regs, int is_fixup)
{
        return (is_fixup ? MIPS_BE_FIXUP : MIPS_BE_FATAL);
}

static int __init spi_gpio_mcp2515_init(void) {
    int ret;

	ret = platform_device_register(&spi_gpio_device);
	if (ret) {
		pr_warning("spi_gpio_mcp2515_init: spi gpio register failed: %d\n", ret);
	}

	ret = platform_device_register(&lmlsfpga_spi_device1);
    if (ret) {
        pr_warning("spi_gpio_mcp2515_init: lmlsfpga_spi_device1 registration failed: %d\n", ret);
    }

	ret = platform_device_register(&lmlsfpga_spi_device2);
    if (ret) {
        pr_warning("spi_gpio_mcp2515_init: lmlsfpga_spi_device2 registration failed: %d\n", ret);
    }
	
	ret = platform_device_register(&lmlsfpga_spi_device3);
    if (ret) {
        pr_warning("spi_gpio_mcp2515_init: lmlsfpga_spi_device3 registration failed: %d\n", ret);
    }

	ret = platform_device_register(&lmlsfpga_spi_device4);
    if (ret) {
        pr_warning("spi_gpio_mcp2515_init: lmlsfpga_spi_device4 registration failed: %d\n", ret);
    }

	ret = spi_register_board_info(spi_gpio_mcp2515_info, ARRAY_SIZE(spi_gpio_mcp2515_info));
	if (ret) {
		pr_warning("spi_gpio_mcp2515_init: spi gpio board register failed: %d\n", ret);
	}

    return ret;
}
device_initcall(spi_gpio_mcp2515_init);


void __init plat_mem_setup(void)
{
	console_verbose();

#ifdef CONFIG_KGDB
	kgdb_config ();
#endif

	mips_reboot_setup();
	board_be_init = vsc452_be_init;
	board_be_handler = vsc452_be_handler;
	return;
}

