/* 
 * (C) Copyright 2003
 * Thomas.Lange@corelatus.se
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <asm/mipsregs.h>
#include <asm/io.h>
#include <vsc452/vsc452regs.h>
#include <vsc452/vsc_gpio.h>

#define BOOTARGS_MEMSIZE_LENGTH 30

#define GPIO(port, number)	GPIO_PORT0 + (port *0x40) + (number * GPIO_PIN_ADDR_UNIT)
#define PHY_RESET		GPIO(1,11)

#define TIMER_4_INITIAL_COUNT   0x10200180
#define TIMER_4_CURRENT_COUNT   0x10200184
#define TIMER_4_MODE_CONTROL    0x10200190
#define TIMER_4_INT_STATUS      0x10200194
#define PARALLEL_CHIP_SELECT_1_CTRL	0x10200004

void UbootWatchdogEnable(void);
void UbootWatchdogDisable(void);

#define write32(addr, val)	writel(val, 0xa0000000 + (addr))
#define read32(addr)		readl(0xa0000000 + (addr))

GPIO_MAP_TABLE Gpio_Table_LML[MAX_GPIOS] =
{

{GPIO_0, GPIO_0, GPIO_DIR_IN, 0},
{GPIO_1, GPIO_1, GPIO_DIR_IN, 0},
{GPIO_2, GPIO_2, GPIO_DIR_IN, 0},
{GPIO_3, GPIO_3, GPIO_DIR_IN, 0}, /* LML SP_PCM_JTAG_EN_L O tristate */
{GPIO_4, GPIO_4, GPIO_DIR_OUT, 1},
{GPIO_5, GPIO_5, GPIO_DIR_OUT, 1}, /* LML PCM_FRU1_WP O high */
{GPIO_6, GPIO_6, GPIO_DIR_OUT, 1},
{GPIO_7, GPIO_7, GPIO_DIR_OUT, 1},
{GPIO_8, GPIO_8, GPIO_DIR_OUT, 1}, /* LML PCM_FRU0_WP O high */
{GPIO_9, GPIO_9, GPIO_DIR_IN, 0},
{GPIO_10, GPIO_10, GPIO_DIR_IN, 0},
{GPIO_11, GPIO_11, GPIO_DIR_IN, 0},
{GPIO_12, GPIO_12, GPIO_DIR_IN, 0},
{GPIO_13, GPIO_13, GPIO_DIR_IN, 0},
{GPIO_14, GPIO_14, GPIO_DIR_OUT, 1},
{GPIO_15, GPIO_15, GPIO_DIR_OUT, 1}, /* LML JTAG_PGM_EN_N O high */
{GPIO_16, GPIO_16, GPIO_DIR_IN, 0},
{GPIO_17, GPIO_17, GPIO_DIR_OUT, 0},
{GPIO_18, GPIO_18, GPIO_DIR_IN, 0},
{GPIO_19, GPIO_19, GPIO_DIR_IN, 0},
{GPIO_20, GPIO_20, GPIO_DIR_IN, 0},
{GPIO_21, GPIO_21, GPIO_DIR_IN, 0}, /* LML not used */
{GPIO_22, GPIO_22, GPIO_DIR_IN, 0},
{GPIO_23, GPIO_23, GPIO_DIR_IN, 0},
{GPIO_24, GPIO_24, GPIO_DIR_IN, 0},
{GPIO_25, GPIO_25, GPIO_DIR_IN, 0},
{GPIO_26, GPIO_26, GPIO_DIR_IN, 0},
{GPIO_27, GPIO_27, GPIO_DIR_IN, 0},
{GPIO_28, GPIO_28, GPIO_DIR_IN, 0},
{GPIO_29, GPIO_29, GPIO_DIR_IN, 0}, /* LML not used */
{GPIO_30, GPIO_30, GPIO_DIR_IN, 0},
{GPIO_31, GPIO_31, GPIO_DIR_IN, 0},     /* Buell/Bimota JTAG: TDI to SP from CPU */
{GPIO_32, GPIO_32, GPIO_DIR_OUT, 1}, /* LML RST_USB_HOST_L O high */
{GPIO_33, GPIO_33, GPIO_DIR_OUT, 1},
{GPIO_34, GPIO_34, GPIO_DIR_OUT, 1}, /* LML RST_CANa0_N O high */
{GPIO_35, GPIO_35, GPIO_DIR_OUT, 1}, /* LML RST_CANa1_N O high */
{GPIO_36, GPIO_36, GPIO_DIR_OUT, 1}, /* LML RST_CANb0_N O high */
{GPIO_37, GPIO_37, GPIO_DIR_OUT, 1}, /* LML RST_CANb1_N O high */
{GPIO_38, GPIO_38, GPIO_DIR_OUT, 1},
{GPIO_39, GPIO_39, GPIO_DIR_OUT, 0},
{GPIO_40, GPIO_40, GPIO_DIR_IN, 0}, /* LML not used */
{GPIO_41, GPIO_41, GPIO_DIR_IN, 0}, /* LML not used */
{GPIO_42, GPIO_42, GPIO_DIR_IN, 0}, /* LML not used */
{GPIO_43, GPIO_43, GPIO_DIR_IN, 0}, /* LML not used */
{GPIO_44, GPIO_44, GPIO_DIR_OUT, 0}, /* LML SEL_EthSW_NCSI O low */
{GPIO_45, GPIO_45, GPIO_DIR_OUT, 0}, /* LML SEL_NCSI O low */
{GPIO_46, GPIO_46, GPIO_DIR_IN, 0}, /* LML not used */
{GPIO_47, GPIO_47, GPIO_DIR_IN, 0}, /* not used */
{GPIO_48, GPIO_48, GPIO_DIR_OUT, 1},
{GPIO_49, GPIO_49, GPIO_DIR_OUT, 1}, /* LML UCD_RST_N O high */
{GPIO_50, GPIO_50, GPIO_DIR_IN, 1}, /* LML UCD9_ALRT1_N I None */
{GPIO_51, GPIO_51, GPIO_DIR_IN, 0},
{GPIO_52, GPIO_52, GPIO_DIR_OUT, 1},    /* Buell/Bimota JTAG: TDO from SP to CPU */
{GPIO_53, GPIO_53, GPIO_DIR_OUT, 1},    /* Buell/Bimota JTAG: TMS */
{GPIO_54, GPIO_54, GPIO_DIR_OUT, 0},    /* Buell/Bimota JTAG: TCK */
{GPIO_55, GPIO_55, GPIO_DIR_IN, 1}, /* LML UCD9_ALRT2_N I None */
{GPIO_56, GPIO_56, GPIO_DIR_IN, 1}, /* LML UCD9_ALRT3_N I None */
{GPIO_57, GPIO_57, GPIO_DIR_IN, 1}, /* LML UCD9_ALRT4_N I None */
{GPIO_58, GPIO_58, GPIO_DIR_OUT, 0},    /* Buell/Bimota JTAG: TRST */
{GPIO_59, GPIO_59, GPIO_DIR_IN, 1}, /* X_SP_PRDY_N I high */
{GPIO_60, GPIO_60, GPIO_DIR_IN, 0},
{GPIO_61, GPIO_61, GPIO_DIR_IN, 0},
{GPIO_62, GPIO_62, GPIO_DIR_IN, 0},
{GPIO_63, GPIO_63, GPIO_DIR_IN, 0},
{GPIO_64, GPIO_64, GPIO_DIR_OUT, 1},
{GPIO_65, GPIO_65, GPIO_DIR_OUT, 1},
{GPIO_66, GPIO_66, GPIO_DIR_IN, 0},
{GPIO_67, GPIO_67, GPIO_DIR_IN, 1}, /* LML SP_IRQ_N1 I high */
{GPIO_68, GPIO_68, GPIO_DIR_IN, 1}, /* LML SP_IRQ_N2 I high */
{GPIO_69, GPIO_69, GPIO_DIR_IN, 1}, /* LML SP_IRQ_N3 I high */
{GPIO_70, GPIO_70, GPIO_DIR_OUT, 1}, /* LML RSTI_NCSI_PHY_N O high */
{GPIO_71, GPIO_71, GPIO_DIR_IN, 0},
{GPIO_72, GPIO_72, GPIO_DIR_IN, 0},
{GPIO_73, GPIO_73, GPIO_DIR_IN, 0},
{GPIO_74, GPIO_74, GPIO_DIR_IN, 0},
{GPIO_75, GPIO_75, GPIO_DIR_IN, 0},
{GPIO_76, GPIO_76, GPIO_DIR_IN, 0},
{GPIO_77, GPIO_77, GPIO_DIR_OUT, 1},
{GPIO_78, GPIO_78, GPIO_DIR_IN, 0}, /* not used */
{GPIO_79, GPIO_79, GPIO_DIR_IN, 1} /* XDP_PRES_N I high */

};


static unsigned long calc_dram_size(void)
{
	/* Calculate size from the DDR Start Operation (DDRSO) register.
	 * Refer to VSC452 manual section 4.2.1 for more details.
	 * Here, we only care about the row, column and bank sizes.
	 *   BS  - Bank size:              If set use 8 banks, else use 4
	 *   MRA - Maximum row address:    Row address bits minus one
	 *   MCA - Maximum column address: Column address bits minus one
	 *
	 * Example DDRSO value below is for a 16-bit wide, 128 MB part
	 * 
	 * xxxx xxxx xxxx xxxx 1001 1100 xxx1 xxxx
	 *                     '--' '--'    "------- BS  -> 1    banks = 8 = 2^3
	 *                      |    '-------------- MRA -> 12   rows  = 2^13
	 *                      '------------------- MCA -> 9    cols  = 2^10
	 *
	 * (rows * columns * banks * bus_width) / 8 = capacity in bytes
	 *
	 * Bus width is fixed at 16 for VSC452, so given the values above:
	 *     2^(3+13+10)*16/8 = 0x08000000 = 128MB
	 */
	unsigned long ddrso = *((unsigned long*)0xB0200F04);
	unsigned long banks   =  (ddrso & (0x0010)) >>  4;
	unsigned long rows    = ((ddrso & (0x0F00)) >>  8) + 1;
	unsigned long columns = ((ddrso & (0xF000)) >> 12) + 1;

	if(banks) {
		banks = 3;  /* 8 banks, 3 bits needed */
	} else {
		banks = 2;  /* 4 banks, 2 bits needed */
	}

	/* Return DRAM size in bytes */
	return ((unsigned long)(1 << (rows + columns + banks)) * (16 / 8));
}


unsigned long initdram(int board_type)
{
	/* Sdram is setup by assembler code */
	/* If memory could be changed, we should return the true value here */
	return calc_dram_size();
}


/* Initialise GPIO pins */
unsigned int
initialize_gpio (void)
{
	int i;
        const unsigned int PIN_OUTPUT_HIGH = 1;
        const unsigned int PIN_OUTPUT_LOW = 0;
        GPIO_MAP_TABLE *Gpio_Table = Gpio_Table_LML;
	uint32_t GPIO_interrupt_status[MAX_GPIO_PORTS] = {0x1020080C, 0x1020081C, 0x1020082C, 0x1020083C, 0x1020084C}; /* These are register offset addresses */

	for (i=0; i<MAX_GPIOS;i++)
        {
                if (i == GPIO_39) //Console Switch. Don't touch
                {
                        continue;
                }
                if (Gpio_Table[i].Direction == GPIO_DIR_OUT) {
                        if (Gpio_Table[i].Value == 1) {
                                write32((GPIO_PORT0+(i*GPIO_PIN_ADDR_UNIT)), PIN_OUTPUT_HIGH);
                        }
                        else if (Gpio_Table[i].Value == 0) {
                                write32((GPIO_PORT0+(i*GPIO_PIN_ADDR_UNIT)), PIN_OUTPUT_LOW);
                        }
                        else {
                                printf("WARNING: pin %d has a value of %d, which is not in the acceptable range\n",
                                                i, Gpio_Table[i].Value);
                        }
                }
                else{
			write32((GPIO_PORT0+(i*GPIO_PIN_ADDR_UNIT)), INPUT_INTR_DISABLE);
                }
        }

	/* Clear all pending GPIO interrupt */
	for(i=0; i < MAX_GPIO_PORTS; i++)
 	{
 		write32(GPIO_interrupt_status[i], 0xFFFF);
	}

	return 0;
}


int checkboard (void)
{
	int i;

	/* Clear all pending I2C interrupt */
	for (i = 0; i < 6; i++)
	{
		*((unsigned long *)(0xB0200504 + i*0x80)) = 1<<7;
	}
	
       /* Enable the SP access of FPGA registers via Flash Chip select 1 (FSC1)  */
	write32(PARALLEL_CHIP_SELECT_1_CTRL, 0x00000012);

	return 0;
}

short 
ReadJumperConfig(void)
{
	/* Return Path ID */	
	return -1;	
}

/*
 * Do watchdog interrupt disabling - its needed in case of cold reset
 * We are supporting 3,4,7 and 8 Timers.
 *
 */
unsigned int
disable_watchdog (void)
{
       /* Timer 3 */
       *((unsigned long *)0xB0200170) = 0;
       *((unsigned long *)0xB0200174) = 0;

       /* Timer 4 */
       *((unsigned long *)0xB0200190) = 0;
       *((unsigned long *)0xB0200194) = 0;

       /* Timer 7 */
       *((unsigned long *)0xB02001F0) = 0;
       *((unsigned long *)0xB02001F4) = 0;

       /* Timer 8 */
       *((unsigned long *)0xB0200210) = 0;
       *((unsigned long *)0xB0200214) = 0;

       return 0;

}

/*
 * Using timer 4 as Uboot Watchdog timer  and this timer  shall be reused again in Kernel, as SP Kernel Watchdog Timer with timer 3
 *   
 */

void
UbootWatchdogEnable(void)
{
        write32(TIMER_4_INITIAL_COUNT, 0x773593FF);
        write32(TIMER_4_INT_STATUS, 0x01);
        write32(TIMER_4_MODE_CONTROL, 0x03);

        return;
}

void
UbootWatchdogDisable(void)
{
        write32(TIMER_4_INITIAL_COUNT, 0x00);
        write32(TIMER_4_MODE_CONTROL, 0x00);
        write32(TIMER_4_INT_STATUS, 0x00);

        return;
}


#ifdef CFG_READ_MAC_FROM_FRU_SUPPORT

/* 
 * Read Platform from FRU 
 * 
 * Set the platformid as Absolute if string  "Absolut" or "SB_XVIII" ( "PLAT:SB_XXXX;" ) found in FRU.  
 * 
 */
int
parsePlatformidAndSet (unsigned char *FruBuf, char * env_platformid)
{
	
	int i ;
	char ReferenceStr[10];
	char LodiStr[5];
	char MaderaStr[7];
	char LivermoreStr[10];
	
	char platStr[6];
	char TempBuf[9]={0};
	
	memcpy(ReferenceStr, "Reference", 9);
	memcpy(LodiStr, "LODI", 4);
	memcpy(MaderaStr, "MADERA", 6);
	memcpy(LivermoreStr, "LIVERMORE", 6);

	memcpy(platStr, "PLAT:", 5);
	
	for (i=0; i<256-6; i++)
	{
		/* Search platfrom id string "Lodi"  */
		if (0 == memcmp(FruBuf+i, LodiStr, 4))
		{
			strcpy (env_platformid, "Lodi");	
			break;
		}
		
		/* Search platfrom id string "Madera"  */
		if (0 == memcmp(FruBuf+i, MaderaStr, 6))
		{
			strcpy(env_platformid, "Madera");
			break;
		}

		/* Search platfrom id string "Livermore"  */
		if (0 == memcmp(FruBuf+i, LivermoreStr, 9))
		{
			strcpy(env_platformid, "Livermore");	
			break;
		}

		/* Search platfrom id string "Reference" */
		if (0 == memcmp(FruBuf+i, ReferenceStr, 9))
		{
			strcpy(env_platformid, "Reference");
			break;
		} 

		/* Search platfrom id string "PLAT:" */
		if (0 == memcmp(FruBuf+i, platStr, 5))
		{
			/* 
		        *If found, copy ASCII platrom "PLAT:SB_XXXX;"  . Store the value of "SB_XXXX"  in TempBuf array and append "\0"as  
			*                              ^-- starts here 
                       *      to last element of array to make it an string ( for string comparison ) 
			*/
			int j=0; 
			/*
			*	"PLAT:SB_XXXX" 
			*             ^--j starts here [ *(FruBuf+i+5+j) ]
			*
			*/
			for (j=0; j<=8; j++)  
			{	
				/*Store the values TempBuf in untill we found the ascii ";"  */
				if (*(FruBuf+i+5+j) == ';') 
				{
					/* Append last element of array with '\0' to make it as string*/
					TempBuf[j] = '\0'; 
					break;
				}		
				else 
				{
					TempBuf[j] = *(FruBuf+i+5+j);
				}
			}
			
			if ((0 == strncmp(TempBuf, "SB_XX", 6)) || (0 == strncmp(TempBuf, "XX", 3)))
			{
				strcpy (env_platformid, "Lodi");
				break ;
			}

			if (0 == strncmp(TempBuf,"SB_XXI", 7))
			{
				strcpy (env_platformid , "Madera");
				break ;
			}
			if (0 == strncmp(TempBuf, "SB_XXII", 8))
			{
				strcpy (env_platformid, "Livermore");
				break ;
			}
			
		} 
		
	}
	if (i >= 256-6)
	{
		printf("Did not find platfromid in FruBuf\n");

#define myprintable(x) (((x)>=0x20)&&((x)<=0x7E))
		int j;
	    for (i = 0; i < 16; ++i) {
			for (j = 0; j < 16; j++) {
				printf ("%02X ", FruBuf[i * 16 + j]);
			}
			printf (" ");

			for (j = 0; j < 16; j++) {
				unsigned char   ch = (char)FruBuf[i * 16 + j];

				if (myprintable(ch)) {
					printf ("%c", ch);
				} else {
					printf ("-");
				}
			}
			printf ("\n");
		}

		return -1;
	}
	return 0;

}

#endif 

/*
 * Read the platformid  from EEPROM.
 * 
 */
unsigned int 
readplatformid ( char *buf  )
{
	char platformid [256] = "Reference";
	char *env_platformid = NULL;
	
#ifdef CFG_READ_MAC_FROM_FRU_SUPPORT

	/* Get the platformid from FRU EEPROM */
	int retval ;
	extern int ReadFru (unsigned char *);
	unsigned char FruBuf[256] = { 0 };
	char platformStr [256] = { 0 };

	/* Read FRU data */
	retval = ReadFru (FruBuf);

	if ( 0 == retval )
	{
		/* Obtain platformid from FRU data */
		retval = parsePlatformidAndSet (FruBuf,platformStr);
		if ( 0 == retval )
		{
			strcpy (buf,platformStr);	
			return strlen(platformStr);
	      	}	
	}
	else 
	{
		printf("WARNING: platformid could not be parsed from FRU, perhaps it is not programmed \n");
	}

#endif 	

	/* Get  the platformid from uboot environment variable , if  EEPROM dont have the platformid */
	env_platformid = getenv("platformid");
	if ( NULL != env_platformid )
	{
	       if ((0 == strncmp(env_platformid, "Reference", 9) ) ||( 0 == strncmp(env_platformid, "Lodi", 4) )|| (0 == strncmp(env_platformid, "Madera", 6)) || ( 0== strncmp(env_platformid, "Livermore", 9))) 
		{
			strcpy(buf, env_platformid);
			return strlen(env_platformid);
		}	
	}	

	/* Set the default platform ID as Reference , if both EEPROM and uboot environment variable dont have the platformid  */
	strcpy (buf, platformid);
	return strlen(platformid);
}


void platform_specific_funcs(void) {
        char plat_str[256];
        char *bootargs;

        /* If needed, update bootcmd to include memory size
         * If there is currently a "memsize" environmemt variable, do nothing
         * so that the kernel will use that value. This is handy if for some
         * reason we miscalculate the memory size here. We will at least have
         * a way to manually intervene and boot normally.
         */
        if(getenv("memsize") == NULL) {
                bootargs = getenv("bootargs");
                if(strlen(bootargs) + BOOTARGS_MEMSIZE_LENGTH < sizeof(plat_str)) {
                        sprintf(plat_str, "%s memsize=%d", bootargs, calc_dram_size());
                        setenv("bootargs", plat_str);
                }
        }
}
