/*
 * (C) Copyright 2001
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Misc functions
 */
#include <common.h>
#include <command.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/time.h>
#include <vsc452/vsc_gpio.h>

#if (CONFIG_COMMANDS )

typedef enum {
	GET_MODULE_GPIO_PROPERTIES,
	GET_GPIO_DIRECTION,
	SET_GPIO_DIRECTION,
	READ_GPIO,
	WRITE_GPIO,
}eGpioFunc;

typedef Gpio_data_t gpio_ioctl_data;

typedef enum {
	GET_GPIO_DIR,
	SET_GPIO_DIR_IN,
	SET_GPIO_DIR_OUT,
	READ_GPIO_VAL,
	SET_GPIO_HIGH,
	SET_GPIO_LOW,
	PRINT_ALL,
	END_OF_FUNCLIST
}eGpioactions;

/* Prototypes */
inline void gpio_write_reg(unsigned long addr, unsigned long val);
inline unsigned long gpio_read_reg(unsigned long addr );
void gpio_init(void);

static gpio_dev_t  gpio_dev;

int vsc_readgpio(gpio_dev_t *gpio,Gpio_data_t *gpindata);
int vsc_writegpio(gpio_dev_t *gpio,Gpio_data_t *gpindata);
int vsc_getdir(gpio_dev_t *gpio, Gpio_data_t *gpindata);
int vsc_setdir(gpio_dev_t *gpio, Gpio_data_t *gpindata);


/* Initial values */
eGpioactions action = END_OF_FUNCLIST;

static int verbose = 0;

uint32_t GPIOPortBaseOffset[MAX_GPIO_PORTS] = {0x0000,
												0x0040,
												0x0080,
												0x00C0,
												0x0100
											  };

/* GPIO Ports and their Pins mapped to GPIO_0 - GPIO_80 */
GPIO_PORT_PIN_MAPPING GpioPortPinMap[MAX_GPIOS] =
{
{GPIO_0, PORT_0, 0},
{GPIO_1, PORT_0, 1},
{GPIO_2, PORT_0, 2},
{GPIO_3, PORT_0, 3},
{GPIO_4, PORT_0, 4},
{GPIO_5, PORT_0, 5},
{GPIO_6, PORT_0, 6},
{GPIO_7, PORT_0, 7},
{GPIO_8, PORT_0, 8},
{GPIO_9, PORT_0, 9},
{GPIO_10, PORT_0, 10},
{GPIO_11, PORT_0, 11},
{GPIO_12, PORT_0, 12},
{GPIO_13, PORT_0, 13},
{GPIO_14, PORT_0, 14},
{GPIO_15, PORT_0, 15},
{GPIO_16, PORT_1, 0},
{GPIO_17, PORT_1, 1},
{GPIO_18, PORT_1, 2},
{GPIO_19, PORT_1, 3},
{GPIO_20, PORT_1, 4},
{GPIO_21, PORT_1, 5},
{GPIO_22, PORT_1, 6},
{GPIO_23, PORT_1, 7},
{GPIO_24, PORT_1, 8},
{GPIO_25, PORT_1, 9},
{GPIO_26, PORT_1, 10},
{GPIO_27, PORT_1, 11},
{GPIO_28, PORT_1, 12},
{GPIO_29, PORT_1, 13},
{GPIO_30, PORT_1, 14},
{GPIO_31, PORT_1, 15},
{GPIO_32, PORT_2, 0},
{GPIO_33, PORT_2, 1},
{GPIO_34, PORT_2, 2},
{GPIO_35, PORT_2, 3},
{GPIO_36, PORT_2, 4},
{GPIO_37, PORT_2, 5},
{GPIO_38, PORT_2, 6},
{GPIO_39, PORT_2, 7},
{GPIO_40, PORT_2, 8},
{GPIO_41, PORT_2, 9},
{GPIO_42, PORT_2, 10},
{GPIO_43, PORT_2, 11},
{GPIO_44, PORT_2, 12},
{GPIO_45, PORT_2, 13},
{GPIO_46, PORT_2, 14},
{GPIO_47, PORT_2, 15},
{GPIO_48, PORT_3, 0},
{GPIO_49, PORT_3, 1},
{GPIO_50, PORT_3, 2},
{GPIO_51, PORT_3, 3},
{GPIO_52, PORT_3, 4},
{GPIO_53, PORT_3, 5},
{GPIO_54, PORT_3, 6},
{GPIO_55, PORT_3, 7},
{GPIO_56, PORT_3, 8},
{GPIO_57, PORT_3, 9},
{GPIO_58, PORT_3, 10},
{GPIO_59, PORT_3, 11},
{GPIO_60, PORT_3, 12},
{GPIO_61, PORT_3, 13},
{GPIO_62, PORT_3, 14},
{GPIO_63, PORT_3, 15},
{GPIO_64, PORT_4, 0},
{GPIO_65, PORT_4, 1},
{GPIO_66, PORT_4, 2},
{GPIO_67, PORT_4, 3},
{GPIO_68, PORT_4, 4},
{GPIO_69, PORT_4, 5},
{GPIO_70, PORT_4, 6},
{GPIO_71, PORT_4, 7},
{GPIO_72, PORT_4, 8},
{GPIO_73, PORT_4, 9},
{GPIO_74, PORT_4, 10},
{GPIO_75, PORT_4, 11},
{GPIO_76, PORT_4, 12},
{GPIO_77, PORT_4, 13},
{GPIO_78, PORT_4, 14},
{GPIO_79, PORT_4, 15}
};

/*iowrite ****/
inline void gpio_write_reg(unsigned long addr, unsigned long val)
{
	*((volatile uint32_t*)addr) = val;
}	

/*********ioread*************/
inline unsigned long gpio_read_reg(unsigned long addr )
{
	return ((*(volatile uint32_t*)addr));
}


void gpio_vsc_init(void){
   gpio_dev.gpio_addr =  (void *) GPIO_BASE_OFFSET;
}

int  
gpio_vsc_ioctl (uint cmd , Gpio_data_t * gpin_data)
{
	gpio_dev_t *dev;

	dev= &gpio_dev; 


	switch(cmd)
	{
		  /* GPIO Functions - */ 
		case  GET_GPIO_DIRECTION :
			
			if (0!=vsc_getdir(dev,gpin_data)){
			   return -1;
            }
 		 	break;
		
		case  SET_GPIO_DIRECTION :
			
			if (0!=vsc_setdir(dev,gpin_data)){
				return -1;
            }
			break;
			
		case   READ_GPIO :	
			
			if (0!=vsc_readgpio(dev,gpin_data)){
				return -1;
            }
 		      break;
			
		case  WRITE_GPIO :
			
			if (0!=vsc_writegpio(dev,gpin_data)){
				return -1;
            }
			break;
		
			
		 default :
			printf("[gpio] : Invalid ioctl(%d) called! \n", cmd);
			return -1;
		}
		return 0;
}






int   
vsc_readgpio(gpio_dev_t *gpio,Gpio_data_t *gpindata)
{
	volatile uint32_t  Addr;
	volatile uint32_t Value;
	unsigned char PortNumber;
	unsigned char PortPinNumber;
	uint32_t temp32;

	if ( gpindata->PinNum > MAX_GPIOS-1)
	{
		return -1;
	}

	/* Get the Port Number and the pin number within that port */
	PortNumber = GpioPortPinMap[gpindata->PinNum].PortNumber;
	PortPinNumber = GpioPortPinMap[gpindata->PinNum].PortPinNumber;

       /* Get the Address for the pin  */ 
	Addr=(uint32_t) gpio_dev.gpio_addr+GPIOPortBaseOffset[PortNumber]+PortPinNumber*GPIO_PIN_ADDR_UNIT;
	Value=gpio_read_reg(Addr);

      /*Check for the 0th bit  for Gpio pin data to be High or Low*/
	temp32=(Value & 0x00000001);

	/*Passing the vaue to the user space*/  
	if (temp32)
	{
		gpindata->data=1;
	}
	else
	{
		gpindata->data=0;
	}
	return 0;
}

int   
vsc_writegpio(gpio_dev_t *gpio,Gpio_data_t *gpindata)
{
	
	volatile uint32_t  Addr;
	volatile uint32_t Value;
	unsigned char PortNumber;
	unsigned char PortPinNumber;

	if ( gpindata->PinNum > MAX_GPIOS-1)
	{
		return -1;
	}

	/* Get the Port Number and the pin number within that port */
	PortNumber = GpioPortPinMap[gpindata->PinNum].PortNumber;
	PortPinNumber = GpioPortPinMap[gpindata->PinNum].PortPinNumber;

	/*Get the Address for the pin */ 
	Addr=(uint32_t) gpio_dev.gpio_addr+GPIOPortBaseOffset[PortNumber]+PortPinNumber*GPIO_PIN_ADDR_UNIT;

	/*Set Gpio pin to be  high or low based on the user data */
	if (gpindata->data )
	{
		Value= (uint32_t)OUTPUT_HIGH;
	}
	else
	{
		Value=(uint32_t)OUTPUT_LOW;
	}

	gpio_write_reg(Addr,Value);
	return 0;
	
}


int   
vsc_getdir(gpio_dev_t *gpio, Gpio_data_t *gpindata)
{
	volatile uint32_t  Addr;
	volatile uint32_t Value;
	unsigned char PortNumber;
	unsigned char PortPinNumber;
	uint32_t temp32;

	if ( gpindata->PinNum > MAX_GPIOS-1)
	{
		return -1;
	}

	/* Get the Port Number and the pin number within that port */
	PortNumber = GpioPortPinMap[gpindata->PinNum].PortNumber;
	PortPinNumber = GpioPortPinMap[gpindata->PinNum].PortPinNumber;

	/*Get the Address for the pin */ 
	Addr=(uint32_t) gpio_dev.gpio_addr+GPIOPortBaseOffset[PortNumber]+PortPinNumber*GPIO_PIN_ADDR_UNIT;
	Value=gpio_read_reg(Addr);

	/*Check for the 1st bit for Direction  to be Input or Output*/
	temp32=(Value & 0x00000002)>>0x01;

      /*Passing the vaue to the user space*/
     /*Return value =1 ,when  Output and  0, when Input */ 
	if (temp32)
	{
		gpindata->data=0;
	}
	else
	{
		gpindata->data=1;
	}
	return 0;
}


int   
vsc_setdir(gpio_dev_t *gpio, Gpio_data_t *gpindata)
{
	volatile uint32_t  Addr;
	volatile uint32_t Value;
	unsigned char PortNumber;
	unsigned char PortPinNumber;

	if ( gpindata->PinNum > MAX_GPIOS-1)
	{
		return -1;
	}

	/* Get the Port Number and the pin number within that port */
	PortNumber = GpioPortPinMap[gpindata->PinNum].PortNumber;
	PortPinNumber = GpioPortPinMap[gpindata->PinNum].PortPinNumber;

	/*Get the Address for the pin */  
	Addr=(uint32_t) gpio_dev.gpio_addr+GPIOPortBaseOffset[PortNumber]+PortPinNumber*GPIO_PIN_ADDR_UNIT;

	/*Set Gpio pin to be  as Input when 0 is received from user space or Output low when  1 received from user space 
	  or Outpur high when 2 is received from user space */
	if (gpindata->data == GPIO_DIR_OUT_DAT_HIGH)
	{
		Value= (uint32_t)OUTPUT_HIGH;
	}
	else if (gpindata->data == GPIO_DIR_OUT_DAT_LOW)
	{
		Value= (uint32_t)OUTPUT_LOW;
	}
	else
	{
		Value=(uint32_t)INPUT_INTR_DISABLE;
	}
	gpio_write_reg(Addr,Value);
	return 0;
}



void gpio_init(void){
     gpio_vsc_init();
}

/** \file gpioifc.c
    \brief Source for all gpio interface code
*/

/**
 * Gpio_action
 * 
 **/
static int gpio_action(  gpio_ioctl_data *argp, unsigned int command )
{
    int ioctl_ret;
    int retval = -1;
    ioctl_ret = gpio_vsc_ioctl(command, argp);
    if (ioctl_ret == -1)
    {
       retval = -1;
    } 
    else
    {
       retval = (int)argp->data;
    }

    return (retval);
}
/**
 * get_gpio_dir
 *
 **/
int get_gpio_dir( unsigned short gpio_number )
{
    gpio_ioctl_data gpio_arg;

    /* Set the gpio number */
    gpio_arg.PinNum = gpio_number;

    return( gpio_action( &gpio_arg, GET_GPIO_DIRECTION ) );
}


/**
 * set_gpio_dir 
 *
 **/
static int set_gpio_dir( unsigned short gpio_number,
                         unsigned char gpio_direction )
{
    gpio_ioctl_data gpio_arg;

    /* Set the gpio number */
    gpio_arg.PinNum = gpio_number;

    /* Set the gpio direction */
    gpio_arg.data = gpio_direction;

    return( gpio_action( &gpio_arg, SET_GPIO_DIRECTION ) );

}


/**
 * set_gpio_dir_intput
 * set the gpio as input 
 **/

inline int set_gpio_dir_input( unsigned short gpio_number )
{
      return( set_gpio_dir( gpio_number, (unsigned char)0) );
}

/**
 * set_gpio_dir_output
 * set the gpio as output 
 **/
inline int set_gpio_dir_output( unsigned short gpio_number )
{
    return( set_gpio_dir( gpio_number, (unsigned char)1 ) );
}

/**
 * set_gpio_dir_output_high
 * set the gpio as output high
 **/
inline int set_gpio_dir_output_high( unsigned short gpio_number )
{
    return( set_gpio_dir( gpio_number, (unsigned char)GPIO_DIR_OUT_DAT_HIGH ) );
}

/**
 * set_gpio_dir_output_low
 * set the gpio as output low
 **/
inline int set_gpio_dir_output_low( unsigned short gpio_number )
{
    return( set_gpio_dir( gpio_number, (unsigned char)GPIO_DIR_OUT_DAT_LOW ) );
}



/**
 * get_gpio_data
 * read the gpio data
 **/
int get_gpio_data( unsigned short gpio_number )
{
     gpio_ioctl_data gpio_arg;

    /* Set the gpio number */
    gpio_arg.PinNum = gpio_number;

    return( gpio_action( &gpio_arg, READ_GPIO ) );

}

/**
 * set_gpio_data
 * write the gpio data
 **/
static int set_gpio_data( unsigned short gpio_number, unsigned char gpio_data )
{
     gpio_ioctl_data gpio_arg;

    /* Set the gpio number */
    gpio_arg.PinNum = gpio_number;

    /* Set the gpio data */
    gpio_arg.data = gpio_data;

    return( gpio_action( &gpio_arg, WRITE_GPIO ) );
}

/**
 * set_gpio_data_high
 * set as high
 **/
inline int set_gpio_data_high( unsigned short gpio_number )
{
    return( set_gpio_data( gpio_number, (unsigned char)1 ) );
}


/**
 * set_gpio_data_low
 * set as low
 **/
inline int set_gpio_data_low( unsigned short gpio_number )
{
    return( set_gpio_data( gpio_number, (unsigned char)0 ) );
}

static void ShowUsuage ( void )
/*@globals fileSystem@*/
/*@modifies fileSystem@*/
{
   printf ("Gpio Test Tool - Ported from Service Processor.\n");	
   printf( "Usage : gpiotool <gpionumber> <option> [--verbose]\n" );
   printf( "        gpiotool <port> <pin> <option> [--verbose]\n" );
   printf( "\t--get-dir:       \tGet the specified gpio's direction\n" );
   printf( "\t--set-dir-input: \tSet the specified gpio to be an input\n" );
   printf( "\t--set-dir-output:\tSet the specified gpio to be an output\n" );

   printf( "\t--get-data:      \tGet the specified gpio's data\n" );	
   printf( "\t--set-data-high: \tSet the specified gpio high\n" );
   printf( "\t--set-data-low:  \tSet the specified gpio low\n" );
   printf( "\t--print-all:     \tPrint out direction and data for all pins\n" );
   printf( "\t--verbose:       \tVerbose for Debugging messages \n" );
   printf( "\n" );
}

static void  
Verbose ( char * msg )
{
	if (verbose ) printf( "%s\n" , msg );

}


static int process_arguments( int argc, char **argv,
			    unsigned short *gpio_num )
{
    int i = 1;
   *gpio_num = MAX_GPIOS ;
	int port = -1, pin = -1;
    while( i < argc )
    {
        /* Check for a command */
        if( strcmp( argv[ i ], "--get-dir" ) == 0 )  
		{
			action = GET_GPIO_DIR;
		}
		else if( strcmp( argv[ i ], "--set-dir-input" ) == 0 )
		{
			action = SET_GPIO_DIR_IN;
		}
        else if( strcmp( argv[ i ], "--set-dir-output" ) == 0 )
        {
            action = SET_GPIO_DIR_OUT;
        }
        else if( strcmp( argv[ i ], "--get-data" ) == 0 )
        {
	    	action = READ_GPIO_VAL;
		}
        else if( strcmp( argv[ i ], "--set-data-low" ) == 0 )
		{
			action = SET_GPIO_LOW;
		}
		else if( strcmp( argv[ i ], "--set-data-high" ) == 0 )	
		{
			action = SET_GPIO_HIGH;
		}
		else if ( strcmp( argv[ i ], "--print-all" ) == 0 )
		{
			action = PRINT_ALL;
		}
		else if( strcmp( argv[ i ], "--verbose" ) == 0 )
			verbose = 1;
		else
		{
			/* If not a command, it should be a gpio  or port/pin*/
			if ( port == -1 )
			{
				port = simple_strtol( argv[ i ], NULL, 10);
			} else {
				pin = simple_strtol( argv[ i ], NULL, 10);
			}
			
		}
		i++;
    }
	if ((port != -1) && (pin != -1))
	{
		*gpio_num = 16*port + pin;
	} else {
		*gpio_num = port;
	}
    return 0;
}




int do_gpiotool(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
   unsigned short gpionum = 0;
   int Value   = 0;
    	
    process_arguments( argc , argv , &gpionum );
    if ( ((gpionum >= MAX_GPIOS) && (action != PRINT_ALL)) || (END_OF_FUNCLIST == action) )
     {
		ShowUsuage ();
		return 0;	
     }

    gpio_init(); 

    switch ( action )
	{
	case GET_GPIO_DIR:
		Verbose   ("Inside Get Dir\n");
		Value = get_gpio_dir ( gpionum );
		if  ( -1 == Value )
		{
			printf ( "Get Dir Failed \n"); 
			return -1;
		}	
		//printf ( "Dir:%d\n",Value);
		if ( 1 == Value ) 	printf ( "Output Pin\n");
		else if ( 0 == Value ) 	printf ( "Input Pin\n");
		else if ( 2 == Value ) 	printf ( "Open Drain Output Pin\n");
		break;
	case SET_GPIO_DIR_IN:
		Verbose   ("Inside Set as Input \n");
		if  ( -1 == set_gpio_dir_input (gpionum ))
		{
			printf ( "Set Dir in Failed \n"); 
			return -1;
		}	
		Verbose   ("GPIO pin is set as input pin \n");
		break;
	case SET_GPIO_DIR_OUT:
		Verbose   ("Inside Set as Output \n");
		if ( -1 == set_gpio_dir_output_low (gpionum ) )
		{
			printf ( "Set Dir out Failed \n"); 
			return -1;
		}	
		Verbose   ("GPIO pin is set as output pin \n");
		break;		
	
	case READ_GPIO_VAL:
		Verbose   ("Inside Read gpio.\n");
		Value = get_gpio_data (gpionum );
		if ( -1 == Value )
		{
			printf ( "Read Gpio failed\n");
			return -1;
		}
		//printf ( "The Pin value is %d\n", Value );
		if ( 1 == Value ) 	printf ( "Pin is High \n");
		else if ( 0 == Value ) 	printf ( "Pin is Low  \n");
		break;	
		
	case SET_GPIO_HIGH:
		Verbose   ("Inside Set Gpio high \n");
		if ( -1 == set_gpio_data_high  (gpionum ) )
		{
			printf  ( "Set Gpio high failed\n");
			return -1;
		}
		Verbose   ("GPIO pin is set to high \n");
		break;		

	case SET_GPIO_LOW:
		Verbose   ("Inside Set Gpio low \n");
		if ( -1 == set_gpio_data_low (gpionum ) )
		{
			printf  ( "Set Gpio low failed\n");
			return -1;
		}
		Verbose   ("GPIO pin is set to low \n");
		break;
	case PRINT_ALL:
                Verbose   ("Inside Print All \n");
                printf("Pin\tDirection\tData\n");
                int i;
                for(i = 0; i < MAX_GPIOS; i++) {
                        // Pin number
                        printf("%d", i);

                        // Pin Direction
                        Value = get_gpio_dir ( i );
                        if ( 1 == Value ) {
                                printf ("\tOutput");
                        }
                        else if ( 0 == Value ) {
                                printf ("\tInput");
                        }

                        // Pin Data
                        Value = get_gpio_data ( i );
                        if ( 1 == Value ) {
                                printf ( "\t\tHigh");
                        }
                        else if ( 0 == Value ) {
                                printf ("\t\tLow");
                        }

                        printf("\n");
                }
                break;
	default:
		Verbose  ("Invalid Gpio Function Call ");
		break;
	}
	return 0;


	return 0;
}

U_BOOT_CMD(
	gpiotool ,    5,    5,     do_gpiotool,
	"gpiotool   - input/output manipulation\n",
    NULL
);

#endif	/* CONFIG_COMMANDS */
