/*
 * $Id$
 *
 * Copyright 2008, Network Appliance Inc.
 * Author: Jason McMullan <mcmullan@netapp.com>
 *
 */

#ifndef ASM_MIPS_ARCH_VSC452_I2C_H
#define ASM_MIPS_ARCH_VSC452_I2C_H

#define ARRAY_SIZE(x) (sizeof(x) /sizeof((x)[0]))

#define VSC452_I2C_INTERFACES	6
#define VSC452_I2C_FIFO_MAX	80	/* 80 byte FIFO */

#define I2C_BASE(x)	(0x10200500 + (0x80 * (x)))

#define SID		0x00	/* Serial Interface Data */
#define SID_DATA(x)		((x) & 0xff)	/* Data */
#define SID_SCA_MASTER		(1 << 7)	/* Master-only interface */
#define SID_SCA_ADDR(x)		((x) & 0x7f)	/* Slave address of interface */
#define SID_SSCA_ENABLE		(1 << 7)	/* Enable SSCA */
#define SID_SSCA_ADDR(x)	((x) & 0x7f)	/* Secondary Slave address of interface */
#define SID_CMS_NAS		(1 << 7)	/* Don't snoop the bus */
#define SID_CMS_CLK_98KHZ	(0 << 0)	/* Clock Select */
#define SID_CMS_CLK_78KHZ	(1 << 0)	/* Clock Select */
#define SID_CMS_CLK_347KHZ	(2 << 0)	/* Clock Select */
#define SID_CMS_CLK_284KHZ	(3 << 0)	/* Clock Select */
#define SID_CMS_CLK_417KHZ	(4 << 0)	/* Clock Select */
#define SID_CMS_CLK_340KHZ	(5 << 0)	/* Clock Select */
#define SID_CMS_CLK_1042KHZ	(6 << 0)	/* Clock Select */
#define SID_CMS_CLK_852KHZ	(7 << 0)	/* Clock Select */

#define SIC		0x04	/* Serial Interface Control */
	/* Read Mode */
#define SIC_PIN		(1 << 7)		/* Pending Interrupt Not */
#define SIC_SSN		(1 << 6)		/* Start Status Not */
#define SIC_STS		(1 << 5)		/* Stop Status */
#define SIC_BER		(1 << 4)		/* Bus Error */
#define SIC_LRB		(1 << 3)		/* Last Received Bit */
#define SIC_AAS		(1 << 2)		/* Addressed as Slave */
#define SIC_LAB		(1 << 1)		/* Lost Arbitaration */
#define SIC_BBN		(1 << 0)		/* Bus Busy Not */
	/* Write Mode */
#define SIC_ES_SCA	(0 << 4)		/* SID is Slave Compare Address */
#define SIC_ES_SSCA	(7 << 4)		/* SID is Secondary Slave Compare Address */
#define SIC_ES_DTH	(1 << 4)		/* SID is Data Out Temporary Holding Reg */
#define SIC_ES_CMS	(2 << 4)		/* SID is Clock and Mode selection */
#define SIC_ES_DATA	(4 << 4)		/* SID is 8-byte data in/out */
#define SIC_ENI		(1 << 3)		/* Enable Interrupt */
#define SIC_STA		(1 << 2)		/* Send Start */
#define SIC_STO		(1 << 1)		/* Send Stop */
#define SIC_ACK		(1 << 0)		/* Acknowledge */

#define SIBC		0x08	/* Serial Interface Bit Control */
#define SIBC_SISR	(1 << 7)		/* Soft Reset */
#define SIBC_SDIE	(1 << 6)		/* Stop Detected Interrupt Enable */
#define SIBC_SCC	(1 << 1)		/* Serial Clock Control */
#define SIBC_SDC	(1 << 0)		/* Serial Data Control */

#define SIBS		0x0c	/* Serial Interface Bit Status */
#define SIBS_SIIS	(1 << 7)		/* Serial Interface Interrupt Status */
#define SIBS_SDIS	(1 << 6)		/* Stop Detected Interrupt Status */
#define SIBS_SCS	(1 << 1)		/* Serial Clock Status */
#define SIBS_SDS	(1 << 0)		/* Serial Data Status */

#define SIFC		0x10	/* Serial Interface RX FIFO Control */
#define SIFC_SRFE	(1 << 7)		/* Slave RX FIFO Enable */
#define SIFC_CRFP	(1 << 6)		/* Clear RX FIFO Pointers */
#define SIFC_STMIE	(1 << 1)		/* Slave TX Mode Interrupt Enable */
#define SIFC_RFIE	(1 << 0)		/* RX FIFO Interrupt Enable */

#define SIFS		0x14	/* Serial Interface RX FIFO Status */
#define SIFS_RFF	(1 << 3)		/* RX FIFO Full */
#define SIFS_RFE	(1 << 2)		/* RX FIFO Empty */
#define SIFS_STMIS	(1 << 1)		/* Slave Transmit Mode Interrupt Status */
#define SIFS_RFIS	(1 << 0)		/* Receive FIFO Interrupt Status */

#define SIFD		0x18	/* Serial Interface RX FIFO Data */
#define SIFD_FD(x)	((x) & 0xff)		/* FIFO Data */

/*-------------------------------------------------------------------------
	Copyright (C) 2008-20013 American Megatrends Inc.
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/

#define STATUS_SUCCESS                          0
#define STATUS_INVALID_PARAM                    -1
#define STATUS_INACCESSIBLE                     -2
#define STATUS_OPER_INCOMPLETE                  -3
#define STATUS_FAILURE                          -4

#define NO_ERROR                                0x00
#define CLEARABLE_ERROR_MASK                    0x1f

 /* PARALLEL_BUS_MLBR */
#define MLBR_LOC_BUS_NUM						0x07
#define MLBR_LOC_BUS_REQ						0x10
#define MLBR_LOC_BUS_ACT						0x20
#define MLBR_LOC_ACC_TYPE						0x40

/* PARALLEL_BUS_MTRR */
#define MTRR_REST_SRT_TIMER						0x01

/* PARALLEL_BUS_MRBR */
#define MRBR_RM_BUS_NUM							0x03
#define MRBR_RM_BUS_GRT							0x10
#define MRBR_RM_BUS_ACT							0x20
#define MRBR_RM_ACC_TYPE						0x40

/* PARALLEL_BUS_MLER */
#define MLER_LOC_REL_SHORT_TMO					0x01
#define MLER_LOC_REL_LONG_TMO					0x02
#define MLER_RM_GRT_TMO							0x04
#define MLER_RM_REL_TMO							0x08
#define MLER_LOC_SER_EXCH_ERR					0x10
#define MLER_LOC_INV_SER_EXCH_BUS				0x20
#define MLER_LOC_BUF_N_READY					0x40

/* PARALLEL_BUS_MTRR */
#define MTRR_START_STIMER						0x01

/* PARALLEL_BUS_MRER */
#define MRER_RM_REL_SHORT_TMO					0x01
#define MRER_RM_REL_LONG_TMO					0x02
#define MRER_LOC_GRT_TMO						0x04
#define MRER_LOC_REL_TMO						0x08
#define MRER_RM_SER_EXCH_ERR					0x10
#define MRER_RM_INV_SER_EXCH_BUS				0x20
#define MRER_RM_BUF_N_READY						0x40

#define SUPFPGA_BASE             0xB6000000
#define SUPFPGA_MCSR_REG         0x1002
#define SUPFPGA_MLBR_REG         0x1006
#define SUPFPGA_MTRR_REG         0x100A
#define SUPFPGA_MRBR_REG         0x100E
#define SUPFPGA_MLER_REG         0x1012
#define SUPFPGA_MRER_REG         0x1016
#define SUPFPGA_MSTR_REG         0x1022
#define SUPFPGA_MLTR_REG         0x1026

/* for support fpga debugging */
#define SUPFPGA_DEBUG0_REG       0xC002
#define SUPFPGA_DEBUG1_REG       0xC006
#define SUPFPGA_DEBUG2_REG       0xC00A
#define SUPFPGA_DEBUG3_REG       0xC00E
#define SUPFPGA_DEBUG4_REG       0xC012
#define SUPFPGA_DEBUG5_REG       0xC016

#define READ_SUPFPGA_MMAP_REG(reg)        (*((uint16_t *)(SUPFPGA_BASE + reg)))
#define WRITE_SUPFPGA_MMAP_REG(reg, val)  *((uint16_t *)(SUPFPGA_BASE + reg)) = (uint16_t)val

#define RESTART_SHORTTIMER

#define RESTART_TIMER_VALUE	         0x0001

#define RESTART_SHORT_TIMER(val)                   \
do {                                               \
    uchar mlbr;                                       \
    supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &mlbr); \
    if (mlbr & MLBR_LOC_BUS_ACT) {                 \
        fpga_start_short_timer();                  \
    }                                              \
} while(0);

#define MAX_SEG_VAL                  7

int i2c_request_shared_bus(uchar segnum);
int i2c_grant_shared_bus(uchar segnum);
int i2c_release_shared_bus(uchar segnum);

int fpga_start_short_timer(void);
int fpga_set_local_short_timeout(uchar timeout);
int fpga_set_local_long_timeout(uchar timeout);

int vsc452_i2c_init(int bus_no);

#endif /* ASM_MIPS_ARCH_VSC452_I2C_H */

